/* DCTC - a Direct Connect text clone for Linux
 * Copyright (C) 2001 Eric Prevoteau
 *
 * he3.c: Copyright (C) Eric Prevoteau <www@ac2i.tzo.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
$Id: he3.c,v 2.3 2003/06/22 06:55:08 blusseau Exp $
*/

static char* HE3_VERSION = "1.01";

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <glib.h>
#include <ctype.h>
#ifdef HAVE_GETOPT_LONG
	#include <getopt.h>
#endif

#include "he3.h"
#include "mmap_tools.h"

char* progname;       /* program name */
char* inputfile = NULL;
char* outputfile = NULL;

/******************************************************/
/* patch from David Toth for                          */
/* File sharing list Windows (DC) compatibiliy bugfix */
/* (byte 4 meaning and computation).                  */
/******************************************************/

/*********************************************************************/
/* When the file list of another user is retrieved, it is compressed */
/* The data format is described below.                               
byte 0: 'H'
byte 1: 'E'
byte 2: '3'
byte 3: 0xD
byte 4: 8 bit parity of the decoded data (the result of all bytes "xor"-ed)
byte 5 - 8 : it is a number encoded in little endian. It it the number of bytes to produce.
byte 9 - 10: it is a number encoded in little endian. This number gives the number of couples.
             Let's name this number as N
byte 11 - (10+2*N): this is a list of couple. The first byte is a character (the one before compression),
             the second byte gives the length of the compressed pattern. Let's name this value L.
byte (11+2*N) - (11+2*N+ ((sum(L)+7)&~7): a bit barbare :) To simplify. Let's name K, the sum of all L values.
             Here, you find a string of K bits (byte aligned). The L0 first bits are the encoded value of 
             caracter of couple 0. The next Lx bits are the encoded value of caractere of couple x.
             Note: each byte are used from the lower bit to the upper bit.
after this string, you have the encoded string.
When data are finally decoded. The format is always:
directory name \xD\xA
\x9 filename | filesize \xD\xA
\x9 filename | filesize \xD\xA
\x9 filename | filesize \xD\xA
\x9 filename | filesize \xD\xA
\x9 filename | filesize \xD\xA

*/

/* ============================================================================================= */
/* ============================================================================================= */
/* ==================================== Decoding functions ===================================== */
/* ============================================================================================= */
/* ============================================================================================= */

/******************************************************/
/*get 1 bit from the current bit position inside data */
/******************************************************/
static unsigned long get_bit(unsigned char *data, unsigned long *cur_pos)
{
	unsigned long out;

	out=((unsigned long)(data[(*cur_pos)/8]>>((*cur_pos)&7)))&1;

	(*cur_pos)++;

	return out;
}

/*********************************************************/
/* get nb_bits from the current bit position inside data */
/*********************************************************/
static unsigned long get_bits(unsigned char *data, unsigned long *cur_pos, int nb_bit)
{
	int i;
	unsigned long res=0;

	for(i=0;i<nb_bit;i++)
	{
		res=(res<<1)|get_bit(data,cur_pos);
	}
	return res;
}

/**************************************************/
/* decompress data compressed using HE3 algorithm */
/**********************************************************/
/* input: a GByteArray containing HE3 compressed data     */
/* output: a GString containing uncompressed data or NULL */
/**********************************************************/
GString *decode_he3_data(unsigned char *data)
{
	GString *output;

	output=g_string_new("");

	if((data[0]=='H')&&(data[1]=='E')&&(data[2]=='3')&&(data[3]==0xD)) {
		GByteArray *decode_array=NULL;
		int pos;
		int nb_couple;
		int max_len=0;		/* max size of encoded pattern */
		int ttl_len=0;		/* total size of all encoded patterns */
		unsigned long offset_pattern;
		unsigned long offset_encoded;
		int nb_output;

		/* compute the number of bytes to produce */
		nb_output=(((int)(data[8]))&255);
		nb_output<<=8;
		nb_output|=(((int)(data[7]))&255);
		nb_output<<=8;
		nb_output|=(((int)(data[6]))&255);
		nb_output<<=8;
		nb_output|=(((int)(data[5]))&255);

		/* compute the number of couples */
		nb_couple=data[9];
		nb_couple+=((((int)(data[10]))&255)<<8);

		for(pos=0;pos<nb_couple;pos++)
		{
			int v;

			v=(((int)(data[12+pos*2]))&255);
			if(v>max_len)
				max_len=v;
			ttl_len+=v;
		}

		decode_array=g_byte_array_new();
		decode_array=g_byte_array_set_size(decode_array,1<<(max_len+1));		/* theorytically, max_len could reach up to 255 */
																								/* but really, a  value above 15 is not oftenly encountered */
																								/* thus the algorithm needs no more than ... 64KB */
																								/* raisonnable value is upto 23 (requires 8MB) */
		{
			GString *tmp;

			tmp=g_string_new("");
			g_string_sprintf(tmp,"he3:max_len: %d",max_len);

			g_string_free(tmp,TRUE);
		}

		if(decode_array!=NULL)
		{
			/* clear the decode array */
			/* the decode array is technically a binary tree */
			/* if the depth of the tree is big (let's say more than 10), */
			/* storing the binary tree inside an array becomes very memory consumming */
			/* but I am too lazy to program all binary tree creation/addition/navigation/destruction functions :) */
			memset(decode_array->data,0,1<<(max_len+1));
			
			offset_pattern=8*(11+nb_couple*2);		/* position of the pattern block, it is just after the list of couples */
			offset_encoded=offset_pattern + ((ttl_len+7)&~7);	/* the encoded data are just after */
																				/* the pattern block (rounded to upper full byte) */

			/* decode_array is a binary tree. byte 0 is the level 0 of the tree. byte 2-3, the level 1, byte 4-7, the level 2, */
			/* in decode array, a N bit length pattern having the value K is its data at the position: */
			/* 2^N + (K&((2^N)-1)) */
			/* due to the fact K has always N bit length, the formula can be simplified into: */
			/* 2^N + K */
			for(pos=0;pos<nb_couple;pos++)
			{
				unsigned int v_len;
				unsigned long value;

				v_len=(((int)(data[12+pos*2]))&255);	/* the number of bit required */

				value=get_bits(data,&offset_pattern,v_len);
				decode_array->data[(1<<v_len)+ value]=data[11+pos*2];		/* the character */
			}

			/* now, its time to decode */
			while(output->len!=nb_output)
			{
				unsigned long cur_val;
				unsigned int nb_bit_val;

				cur_val=get_bit(data,&offset_encoded);		/* get one bit */
				nb_bit_val=1;

				while(decode_array->data[(1<<nb_bit_val) + cur_val ]==0)
				{
					cur_val=(cur_val<<1)|get_bit(data,&offset_encoded);
					nb_bit_val++;
				}
				
				output=g_string_append_c(output,decode_array->data[(1<<nb_bit_val) + cur_val ]);
			}

			g_byte_array_free(decode_array,TRUE);
		}
	} else {
		fprintf(stderr,"Error: the input file is not a HE3 encoded file !");
	}
	return output;
}

/* ============================================================================================= */
/* ============================================================================================= */
/* ==================================== Encoding functions ===================================== */
/* ============================================================================================= */
/* ============================================================================================= */
typedef struct hufnode
{
	unsigned long occur;
	struct hufnode *left,*right;		/* son nodes (left and right are both either not null or either null) */
	unsigned char val;					/* if left&right!=NULL, val is the encoded character, else, it has no usage */
} HUFNODE;

typedef struct
{
	unsigned int bits_len;				/* number of bits in the bitfield used to encode the value */
	unsigned long bits;					/* bitfield containing the encoding pattern */
												/* a N bits_len pattern is stored inside bits from bit N-1 to bit 0 */
} HUFENCODE;

static gint huf_insert_glist(gconstpointer a, gconstpointer b)
{
	if(((const HUFNODE*)a)->occur<((const HUFNODE*)b)->occur)
		return -1;		/* a is must be before b */
	else if(((const HUFNODE*)a)->occur>((const HUFNODE*)b)->occur)
			return 1;		/* a is must be after b */

	/* if both have the same occurences, the one without son comes before */
	if( (((const HUFNODE*)a)->left==NULL) && (((const HUFNODE*)b)->left==NULL) )
		return -1;
	
	if(((const HUFNODE*)a)->left==NULL)
		return -1;

	return 1;
}

/**************************************************************************/
/* recursively scan all nodes of the huffman tree and fill encoding table */
/**************************************************************************/
static void use_hufnode(HUFENCODE tbl_enc[256], HUFNODE *node,unsigned int bits_len, unsigned long bits)
{
	if(node->left!=NULL)
	{	/* it is a node, right is always != NULL */
		use_hufnode(tbl_enc,node->left,bits_len+1,(bits<<1)|0);
		use_hufnode(tbl_enc,node->right,bits_len+1,(bits<<1)|1);
	}
	else
	{	/* it is a leaf, right is always NULL */
		int idx=((int)node->val)&255;
		tbl_enc[idx].bits_len=bits_len;
		tbl_enc[idx].bits=bits;
#if 0
		printf("huf: bits_len: %u  pattern: %lu ",bits_len,bits);
		printf("leaf: %d (%c)\n",idx,node->val);
#endif

		if(bits_len>8*sizeof(unsigned long))		/* compared with the size of unsigned long in bits */
			fprintf(stderr,"user_hufnode: encoded value cannot fit into unsigned long !");
	}
	return;
}

/*****************************************************************/
/* recursively free memory used by all nodes of the huffman tree */
/*****************************************************************/
static void free_hufnode(HUFNODE *node)
{
	if(node==NULL)
		return;
	if(node->left!=NULL)
		free_hufnode(node->left);
	if(node->right!=NULL)
		free_hufnode(node->right);
	free(node);
}


static GByteArray *add_bit(GByteArray *data, unsigned long *bit_pos, unsigned char bit_value)
{
	if(((*bit_pos)&7)==0)		/* starting a new byte ? */
	{
		unsigned char v=0;
		data=g_byte_array_append(data,&v,1);
	}

	/* due to the fact we always add 0 as new byte, we don't have to set bit having 0 as value */
	/* but just 1 */
	if(bit_value!=0)
	{
		data->data[(*bit_pos)/8]|= (1<<((*bit_pos)&7));
	}

	(*bit_pos)++;

	return data;
}

/**************************************/
/* append the pattern to data@bit_pos */
/**************************************/
static GByteArray *add_bits(GByteArray *data, unsigned long *bit_pos, unsigned long pattern, unsigned int pattern_length)
{
	int i;

	for(i=0;i<pattern_length;i++)
	{
		data=add_bit(data,bit_pos,	(pattern>>(pattern_length-1-i))&1 );		/* use pattern from upper to lower bit */
	}
	return data;
}

#if 0
static void	disp_huf(GList *pre_tree)
{
	int i;
	HUFNODE *w;
	printf("---\n");
	for(i=0;i<g_list_length(pre_tree);i++)
	{
		w=g_list_nth_data(pre_tree,i);		/* get the first HUFNODE of the list */

		printf("occur: %ld, left=%p, right=%p, val=%02X\n",w->occur,w->left,w->right,w->val);
	}
}
#endif

/*****************************************************************************************/
/* compress data compressed using an Huffman algorithm and store it in HE3 usable format */
/*****************************************************************************************/
/* input: a GString containing a string to compress        */
/* output: a GByteArray containing compressed data or NULL */
/***********************************************************/
GByteArray *encode_he3_data(GString *str)
{
	unsigned long occur[256];
	HUFENCODE tbl_enc[256];

	int i;
	GByteArray *data;
	GList *pre_tree=NULL;
	HUFNODE *root_huf;
	int nb_val=0;
	unsigned long bit_pos;

	if((str==NULL)||(str->len==0))
		return NULL;

	/* count the number of times each character appears */
	memset(occur,0,sizeof(occur));

	for(i=0;i<str->len;i++)
	{
		occur[((int)(str->str[i]))&255]++;
	}

	/* now, we will build the huffman tree */

	/* stage 1: create all the leafs */
	for(i=0;i<256;i++)
	{
		if(occur[i]!=0)
		{
			HUFNODE *nw;

			nw=malloc(sizeof(HUFNODE));
			nw->occur=occur[i];
			nw->left=NULL;
			nw->right=NULL;
			nw->val=i;

			pre_tree=g_list_insert_sorted(pre_tree,nw, huf_insert_glist);

			nb_val++;
		}
	}

	while(g_list_length(pre_tree)>1)
	{
		HUFNODE *nw;

#if 0
		disp_huf(pre_tree);
#endif
		nw=malloc(sizeof(HUFNODE));
		nw->left=g_list_nth_data(pre_tree,0);		/* get the first HUFNODE of the list */
		pre_tree=g_list_remove(pre_tree,nw->left);/* and remove it */
		nw->right=g_list_nth_data(pre_tree,0);		/* get the second HUFNODE of the list */
		pre_tree=g_list_remove(pre_tree,nw->right);/* and remove it */
		nw->occur=nw->left->occur+nw->right->occur;
		nw->val=0;

		/* add a new node having the 2 removed nodes as son */
		pre_tree=g_list_insert_sorted(pre_tree,nw, huf_insert_glist);
	}
#if 0
	disp_huf(pre_tree);
#endif

	/* now, the huffman tree is done */
	root_huf=g_list_nth_data(pre_tree,0);
	pre_tree=g_list_remove(pre_tree,root_huf);

	memset(tbl_enc,0,sizeof(tbl_enc));

	/* now, we will compute encoding pattern from huffman tree */
	use_hufnode(tbl_enc,root_huf,0,0);

	/* well, encoding table is now done, we can encode the data */
	data=g_byte_array_new();

	/* set the initial HE3 header */
	{
		unsigned char he3_header[]={'H','E','3',0xD,
											 0,					/* parity: all uncompressed bytes xor'ed */
											 0,0,0,0,			/* number of bytes to produce (little-endian) */
											 0,0};				/* number of couples */

		/* patch from David Toth (byte 4: parity computation) */
		unsigned char parity=0;

		for(i=0;i<str->len;i++)
			parity^=str->str[i];
		he3_header[4]=parity&255;

		he3_header[5]=str->len&255;
		he3_header[6]=(str->len>>8)&255;
		he3_header[7]=(str->len>>16)&255;
		he3_header[8]=(str->len>>24)&255;
		he3_header[9]=nb_val&255;
		he3_header[10]=(nb_val>>8)&255;

		data=g_byte_array_append(data,he3_header,11);
	}

	/* add the couple list (character, pattern length) */
	for(i=0;i<256;i++)
	{
		if(occur[i]!=0)
		{
			unsigned char ent[2];

			ent[0]=i;
			ent[1]=tbl_enc[i].bits_len;
			data=g_byte_array_append(data,ent,2);
		}
	}

	/* and now, add all the patterns */
	bit_pos=data->len*8;
	for(i=0;i<256;i++)
	{
		if(occur[i]!=0)
		{
			data=add_bits(data,&bit_pos,tbl_enc[i].bits, tbl_enc[i].bits_len);
		}
	}

	bit_pos=(bit_pos+7)&~7;		/* move to the beginning of the next byte */

	/* now, we encode the string */
	for(i=0;i<str->len;i++)
	{
		int idx=((int)str->str[i])&255;

		data=add_bits(data,&bit_pos,tbl_enc[idx].bits, tbl_enc[idx].bits_len);
	}

	/* free huffman tree */
	free_hufnode(root_huf);

	return data;
}

#ifdef WIN32 /* Windows NT */
# 	 define PATH_SEP2 '\\'
#  	define PATH_SEP3 ':'
#	define casemap(c) tolow(c) /* Force file names to lower case */
#endif

#ifndef PATH_SEP
#  	define PATH_SEP '/'
#endif

#ifndef casemap
#  	define casemap(c) (c)
#endif

#define tolow(c)  (isupper (c) ? tolower (c) : (c))  /* force to lower case */

/* ========================================================================
 * Put string s in lower case, return s.
 */
char *strlwr(s)
    char *s;
{
    char *t;
    for (t = s; *t; t++)
      *t = tolow ((unsigned char) *t);
    return s;
}

/* ========================================================================
 * Return the base name of a file (remove any directory prefix and
 * any version suffix). For systems with file names that are not
 * case sensitive, force the base name to lower case.
 */
char *base_name(fname)
    char *fname;
{
    char *p;

    if ((p = strrchr(fname, PATH_SEP))  != NULL) fname = p+1;
#ifdef PATH_SEP2
    if ((p = strrchr(fname, PATH_SEP2)) != NULL) fname = p+1;
#endif
#ifdef PATH_SEP3
    if ((p = strrchr(fname, PATH_SEP3)) != NULL) fname = p+1;
#endif
#ifdef SUFFIX_SEP
    if ((p = strrchr(fname, SUFFIX_SEP)) != NULL) *p = '\0';
#endif
    if (casemap('A') == 'a') strlwr(fname);
    return fname;
}

/* ======================================================================== */
void usage() {
    printf ("usage: %s [-h] [-o FILE] file\n", progname);
}

/* ======================================================================== */
void help() {
    static char  *help_msg[] = {
		" -h --help        give this help",
		" -o --output=FILE write to FILE instead of stdout",
		" file             file to decompress.",
		0};
    char **p = help_msg;

    printf ("%s %s\n", progname, HE3_VERSION);
    usage();
    while (*p) printf ("%s\n", *p++);
}

static void parse_command_line_options(int argc,char **argv) {

#ifdef HAVE_GETOPT_LONG
	static struct option optab[]= {
		{"help",no_argument,NULL,'h'},          /* get help */
		{"output",required_argument,NULL,'o'},		/* output to file */
		{NULL,0,NULL,'\0'}               		/* last option */
	};
#endif
	static const char *short_opt="ho:";

	int ch; 

	if(argc == 1) {
		usage();
		exit(1);
	}
	opterr = 0;
#ifdef HAVE_GETOPT_LONG
	while((ch=getopt_long(argc,argv,short_opt,optab,NULL))!=EOF)
#else
	while((ch=getopt(argc,argv,short_opt))!=EOF)
#endif
	{
		switch(ch) {
			case 'h':
				help(); exit(0); break;
			case 'o':
				outputfile=strdup(optarg); break;
			default:
				fprintf(stderr,"Unknown option: %c\n",ch);
				usage();
				exit(1);
		}
	}

	if (argv[optind] == NULL) {
		fprintf (stderr,"You need to supply a file for decompression !\n");
		usage();
		exit(1);
	}
	inputfile=strdup(argv[optind]);
}

int main(int argc,char **argv) {
	unsigned char* in;
	GString *out;
	int filesize;
	FILE* foutput=NULL;
	
	progname = base_name (argv[0]);

	parse_command_line_options(argc,argv);

	if (outputfile) {
		foutput=fopen(outputfile,"wb");
		if (foutput == NULL) {
			fprintf(stderr,"%s: can't create file %s (%s)\n",progname,outputfile,strerror(errno));
			exit(1);
		}
	}

	in=map_file(inputfile,&filesize);
	
	if (in != NULL) {
		out=decode_he3_data(in);
		unmap_file(in,filesize);
		if (foutput)
			fprintf(foutput,"%s",out->str);
		else
			printf("%s",out->str);
	}
	exit(0);
}
